/*******************************************************************
 * This file is part of the Emulex Linux Device Driver for         *
 * Enterprise Fibre Channel Host Bus Adapters.                     *
 * Refer to the README file included with this package for         *
 * driver version and adapter support.                             *
 * Copyright (C) 2004 Emulex Corporation.                          *
 * www.emulex.com                                                  *
 *                                                                 *
 * This program is free software; you can redistribute it and/or   *
 * modify it under the terms of the GNU General Public License     *
 * as published by the Free Software Foundation; either version 2  *
 * of the License, or (at your option) any later version.          *
 *                                                                 *
 * This program is distributed in the hope that it will be useful, *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the   *
 * GNU General Public License for more details, a copy of which    *
 * can be found in the file COPYING included with this package.    *
 *******************************************************************/

/*
 * $Id: lpfc_scsiport.c 1.153 2004/09/20 20:05:01EDT sf_support Exp  $
 */
#include <linux/version.h>
#include <linux/spinlock.h>
#include <linux/pci.h>
#include <linux/blkdev.h>
#include <scsi/scsi.h>
#include <scsi/scsi_device.h>
#include <scsi/scsi_host.h>
#include <scsi/scsi_cmnd.h>
#include "lpfc_hw.h"
#include "lpfc_sli.h"
#include "lpfc_mem.h"
#include "lpfc_disc.h"
#include "lpfc_scsi.h"
#include "lpfc.h"
#include "lpfc_logmsg.h"
#include "lpfc_fcp.h"
#include "lpfc_crtn.h"
#include "lpfc_compat.h"

typedef struct lpfc_xlat_err {
	uint16_t iocb_status;
	uint16_t host_status;
	uint16_t action_flag;
} lpfc_xlat_err_t;

/* Defines for action flags */
#define LPFC_FCPRSP_ERROR    0x2
#define LPFC_IOERR_TABLE     0x4
#define LPFC_STAT_ACTION     0x8

#define LPFC_CMD_BEING_RETRIED  0xFFFF

/* This table is indexed by the IOCB ulpStatus */

static lpfc_xlat_err_t lpfc_iostat_tbl[IOSTAT_CNT] = {
/* f/w code            host_status   flag */

	{IOSTAT_SUCCESS, DID_OK, 0},	/* 0x0 */
	{IOSTAT_FCP_RSP_ERROR, DID_OK, LPFC_FCPRSP_ERROR},	/* 0x1 */
	{IOSTAT_REMOTE_STOP, DID_ERROR, 0},	/* 0x2 */
	{IOSTAT_LOCAL_REJECT, DID_ERROR, 0},	/* 0x3 */
	{IOSTAT_NPORT_RJT, DID_ERROR, LPFC_STAT_ACTION},	/* 0x4 */
	{IOSTAT_FABRIC_RJT, DID_ERROR, LPFC_STAT_ACTION},	/* 0x5 */
	{IOSTAT_NPORT_BSY, DID_BUS_BUSY, 0},	/* 0x6 */
	{IOSTAT_FABRIC_BSY, DID_BUS_BUSY, 0},	/* 0x7 */
	{IOSTAT_INTERMED_RSP, DID_ERROR, 0},	/* 0x8 */
	{IOSTAT_LS_RJT, DID_ERROR, 0},	/* 0x9 */
	{IOSTAT_BA_RJT, DID_ERROR, 0},	/* 0xa */
	{IOSTAT_RSVD1, DID_ERROR, 0},	/* 0xb */
	{IOSTAT_RSVD2, DID_ERROR, 0},	/* 0xc */
	{IOSTAT_RSVD3, DID_ERROR, 0},	/* 0xd */
	{IOSTAT_RSVD4, DID_ERROR, 0},	/* 0xe */
	{IOSTAT_DEFAULT, DID_ERROR, 0},	/* 0xf */
	{IOSTAT_DRIVER_REJECT, DID_ERROR, 0}	/* 0x10 */
};


/* Functions required by the scsiport module. */

/* This routine allocates a scsi buffer, which contains all the necessary
 * information needed to initiate a SCSI I/O. The non-DMAable region of
 * the buffer contains the area to build the IOCB. The DMAable region contains
 * the memory for the FCP CMND, FCP RSP, and the inital BPL.
 * In addition to allocating memeory, the FCP CMND and FCP RSP BDEs are setup
 * in the BPL and the BPL BDE is setup in the IOCB.
 */
static struct lpfc_scsi_buf *
lpfc_get_scsi_buf(struct lpfc_hba * phba, int gfp_flags)
{
	struct lpfc_scsi_buf *psb;
	struct ulp_bde64 *bpl;
	IOCB_t *cmd;
	uint8_t *ptr;
	dma_addr_t pdma_phys;

	if ((psb = mempool_alloc(phba->scsibuf_mem_pool, gfp_flags)) == 0) {
		return (0);
	}
	memset(psb, 0, sizeof (struct lpfc_scsi_buf));

	/* Get a SCSI DMA extention for an I/O */
	/*
	 * The DMA buffer for struct fcp_cmnd, struct fcp_rsp and BPL use
	 * lpfc_scsi_dma_ext_pool with size LPFC_SCSI_DMA_EXT_SIZE
	 *
	 *
	 *    The size of struct fcp_cmnd  = 32 bytes.
	 *    The size of struct fcp_rsp   = 160 bytes.
	 *    The size of struct ulp_bde64 = 12 bytes and driver can only
	 *    support LPFC_SCSI_INITIAL_BPL_SIZE (3) S/G segments for scsi data.
	 *    One struct ulp_bde64 is used for each of the struct fcp_cmnd and
	 *    struct fcp_rsp
	 *
	 *    Total usage for each I/O use 32 + 160 + (2 * 12) +
	 *    (4 * 12) = 264 bytes.
	 */

	INIT_LIST_HEAD(&psb->dma_ext.list);

	psb->dma_ext.virt = pci_pool_alloc(phba->lpfc_scsi_dma_ext_pool,
					   GFP_ATOMIC, &psb->dma_ext.phys);
	if (!psb->dma_ext.virt) {
		kfree(psb);
		return (0);
	}

	/* Save virtual ptrs to FCP Command, Response, and BPL */
	ptr = (uint8_t *) psb->dma_ext.virt;

	memset(ptr, 0, LPFC_SCSI_DMA_EXT_SIZE);
	psb->fcp_cmnd = (struct fcp_cmnd *) ptr;
	ptr += sizeof (struct fcp_cmnd);
	psb->fcp_rsp = (struct fcp_rsp *) ptr;
	ptr += (sizeof (struct fcp_rsp));
	psb->fcp_bpl = (struct ulp_bde64 *) ptr;
	psb->scsi_hba = phba;

	/* Since this is for a FCP cmd, the first 2 BDEs in the BPL are always
	 * the FCP CMND and FCP RSP, so lets just set it up right here.
	 */
	bpl = psb->fcp_bpl;
	/* ptr points to physical address of FCP CMD */
	pdma_phys = psb->dma_ext.phys;
	bpl->addrHigh = le32_to_cpu(putPaddrHigh(pdma_phys));
	bpl->addrLow = le32_to_cpu(putPaddrLow(pdma_phys));
	bpl->tus.f.bdeSize = sizeof (struct fcp_cmnd);
	bpl->tus.f.bdeFlags = BUFF_USE_CMND;
	bpl->tus.w = le32_to_cpu(bpl->tus.w);
	bpl++;

	/* Setup FCP RSP */
	pdma_phys += sizeof (struct fcp_cmnd);
	bpl->addrHigh = le32_to_cpu(putPaddrHigh(pdma_phys));
	bpl->addrLow = le32_to_cpu(putPaddrLow(pdma_phys));
	bpl->tus.f.bdeSize = sizeof (struct fcp_rsp);
	bpl->tus.f.bdeFlags = (BUFF_USE_CMND | BUFF_USE_RCV);
	bpl->tus.w = le32_to_cpu(bpl->tus.w);
	bpl++;

	/* Since the IOCB for the FCP I/O is built into the struct
	 * lpfc_scsi_buf, lets setup what we can right here.
	 */
	pdma_phys += (sizeof (struct fcp_rsp));
	cmd = &psb->cur_iocbq.iocb;
	cmd->un.fcpi64.bdl.ulpIoTag32 = 0;
	cmd->un.fcpi64.bdl.addrHigh = putPaddrHigh(pdma_phys);
	cmd->un.fcpi64.bdl.addrLow = putPaddrLow(pdma_phys);
	cmd->un.fcpi64.bdl.bdeSize = (2 * sizeof (struct ulp_bde64));
	cmd->un.fcpi64.bdl.bdeFlags = BUFF_TYPE_BDL;
	cmd->ulpBdeCount = 1;
	cmd->ulpClass = CLASS3;

	return (psb);
}

/* This routine frees a scsi buffer, both DMAable and non-DMAable regions */
static void
lpfc_free_scsi_buf(struct lpfc_scsi_buf * psb)
{
	struct lpfc_hba *phba = psb->scsi_hba;
	struct lpfc_dmabuf *pbpl, *next_bpl;

	/* Check to see if there were any extra buffers used to chain BPLs */
	list_for_each_entry_safe(pbpl, next_bpl, &psb->dma_ext.list, list) {
		lpfc_mbuf_free(phba, pbpl->virt, pbpl->phys);
		list_del(&pbpl->list);
		kfree(pbpl);
	}
	pci_pool_free(phba->lpfc_scsi_dma_ext_pool, psb->dma_ext.virt,
							 psb->dma_ext.phys);
	mempool_free(psb, phba->scsibuf_mem_pool);

	return;
}

static int
lpfc_os_prep_io(struct lpfc_hba * phba, struct lpfc_scsi_buf * lpfc_cmd)
{
	struct fcp_cmnd *fcp_cmnd;
	struct ulp_bde64 *topbpl;
	struct ulp_bde64 *bpl;
	struct lpfc_dmabuf *bmp;
	struct lpfc_dmabuf *head_bmp;
	IOCB_t *cmd;
	struct scsi_cmnd *cmnd;
	struct scatterlist *sgel_p;
	dma_addr_t physaddr;
	uint32_t seg_cnt, i;
	uint32_t num_bmps, num_bde, max_bde;
	uint16_t use_sg;
	int datadir;

	bpl = lpfc_cmd->fcp_bpl;
	fcp_cmnd = lpfc_cmd->fcp_cmnd;

	bpl += 2;		/* Bump past FCP CMND and FCP RSP */
	max_bde = LPFC_SCSI_INITIAL_BPL_SIZE - 1;

	cmnd = lpfc_cmd->pCmd;
	cmd = &lpfc_cmd->cur_iocbq.iocb;

	/* These are needed if we chain BPLs */
	head_bmp = &(lpfc_cmd->dma_ext);
	num_bmps = 1;
	topbpl = 0;

	use_sg = cmnd->use_sg;
	num_bde = 0;
	sgel_p = 0;

	/*
	 * Fill in the FCP CMND
	 */
	memcpy(&fcp_cmnd->fcpCdb[0], cmnd->cmnd, 16);

	if (cmnd->device->tagged_supported) {
		switch (cmnd->tag) {
		case HEAD_OF_QUEUE_TAG:
			fcp_cmnd->fcpCntl1 = HEAD_OF_Q;
			break;
		case ORDERED_QUEUE_TAG:
			fcp_cmnd->fcpCntl1 = ORDERED_Q;
			break;
		default:
			fcp_cmnd->fcpCntl1 = SIMPLE_Q;
			break;
		}
	} else {
		fcp_cmnd->fcpCntl1 = 0;
	}

	datadir = cmnd->sc_data_direction;
	lpfc_cmd->datadir = cmnd->sc_data_direction;

	/* This next section finishes building the BPL for the I/O from the
	 * scsi_cmnd and updates the IOCB accordingly.
	 */
	if (use_sg) {
		sgel_p = (struct scatterlist *)cmnd->request_buffer;
		seg_cnt = dma_map_sg(&phba->pcidev->dev, sgel_p, use_sg,
				     datadir);

		/* return error if we cannot map sg list */
		if (seg_cnt == 0) {
			return 1;
		}

		/* scatter-gather list case */
		for (i = 0; i < seg_cnt; i++) {
			/* Check to see if current BPL is full of BDEs */
			/* If this is last BDE and there is one left in */
			/* current BPL, use it.                         */
			if (num_bde == max_bde) {
				bmp = kmalloc(sizeof (struct lpfc_dmabuf),
					      GFP_ATOMIC);
				if (bmp == 0) {
					goto error_out;
				}
				memset(bmp, 0, sizeof (struct lpfc_dmabuf));
				bmp->virt =
				    lpfc_mbuf_alloc(phba, 0, &bmp->phys);
				if (!bmp->virt) {
					kfree(bmp);
					goto error_out;
				}
				max_bde = ((1024 / sizeof(struct ulp_bde64))-3);
				/* Fill in continuation entry to next bpl */
				bpl->addrHigh =
				    le32_to_cpu(putPaddrHigh(bmp->phys));
				bpl->addrLow =
				    le32_to_cpu(putPaddrLow(bmp->phys));
				bpl->tus.f.bdeFlags = BPL64_SIZE_WORD;
				num_bde++;
				if (num_bmps == 1) {
					cmd->un.fcpi64.bdl.bdeSize += (num_bde *
						sizeof (struct ulp_bde64));
				} else {
					topbpl->tus.f.bdeSize = (num_bde *
						sizeof (struct ulp_bde64));
					topbpl->tus.w =
					    le32_to_cpu(topbpl->tus.w);
				}
				topbpl = bpl;
				bpl = (struct ulp_bde64 *) bmp->virt;
				list_add(&bmp->list, &head_bmp->list);
				num_bde = 0;
				num_bmps++;
			}

			physaddr = sg_dma_address(sgel_p);

			bpl->addrLow = le32_to_cpu(putPaddrLow(physaddr));
			bpl->addrHigh = le32_to_cpu(putPaddrHigh(physaddr));
			bpl->tus.f.bdeSize = sg_dma_len(sgel_p);
			if (datadir == DMA_TO_DEVICE)
			{
				bpl->tus.f.bdeFlags = 0;
			} else {
				bpl->tus.f.bdeFlags = BUFF_USE_RCV;
			}
			bpl->tus.w = le32_to_cpu(bpl->tus.w);
			bpl++;
			sgel_p++;
			num_bde++;
		}		/* end for loop */

		if (datadir == DMA_TO_DEVICE)
		{
			cmd->ulpCommand = CMD_FCP_IWRITE64_CR;
			fcp_cmnd->fcpCntl3 = WRITE_DATA;

			phba->fc4OutputRequests++;
		} else {
			cmd->ulpCommand = CMD_FCP_IREAD64_CR;
			cmd->ulpPU = PARM_READ_CHECK;
			cmd->un.fcpi.fcpi_parm = cmnd->request_bufflen;
			fcp_cmnd->fcpCntl3 = READ_DATA;

			phba->fc4InputRequests++;
		}
	} else {
		if (cmnd->request_buffer && cmnd->request_bufflen) {

			physaddr = dma_map_single(&phba->pcidev->dev,
						  cmnd->request_buffer,
						  cmnd->request_bufflen,
						  datadir);

			/* no scatter-gather list case */
			lpfc_cmd->nonsg_phys = physaddr;
			bpl->addrLow = le32_to_cpu(putPaddrLow(physaddr));
			bpl->addrHigh = le32_to_cpu(putPaddrHigh(physaddr));
			bpl->tus.f.bdeSize = cmnd->request_bufflen;
			if (datadir == DMA_TO_DEVICE)
			{
				cmd->ulpCommand = CMD_FCP_IWRITE64_CR;
				fcp_cmnd->fcpCntl3 = WRITE_DATA;
				bpl->tus.f.bdeFlags = 0;

				phba->fc4OutputRequests++;
			} else {
				cmd->ulpCommand = CMD_FCP_IREAD64_CR;
				cmd->ulpPU = PARM_READ_CHECK;
				cmd->un.fcpi.fcpi_parm = cmnd->request_bufflen;
				fcp_cmnd->fcpCntl3 = READ_DATA;
				bpl->tus.f.bdeFlags = BUFF_USE_RCV;

				phba->fc4InputRequests++;
			}
			bpl->tus.w = le32_to_cpu(bpl->tus.w);
			num_bde = 1;
			bpl++;
		} else {
			cmd->ulpCommand = CMD_FCP_ICMND64_CR;
			cmd->un.fcpi.fcpi_parm = 0;
			fcp_cmnd->fcpCntl3 = 0;

			phba->fc4ControlRequests++;
		}
	}
	bpl->addrHigh = 0;
	bpl->addrLow = 0;
	bpl->tus.w = 0;
	if (num_bmps == 1) {
		cmd->un.fcpi64.bdl.bdeSize +=
			(num_bde * sizeof (struct ulp_bde64));
	} else {
		topbpl->tus.f.bdeSize = (num_bde * sizeof (struct ulp_bde64));
		topbpl->tus.w = le32_to_cpu(topbpl->tus.w);
	}
	cmd->ulpBdeCount = 1;
	cmd->ulpLe = 1;		/* Set the LE bit in the iocb */

	/* set the Data Length field in the FCP CMND accordingly */
	fcp_cmnd->fcpDl = be32_to_cpu(cmnd->request_bufflen);

	return (0);

error_out:
	/* Allocation of a chained BPL failed, unmap the sg list and return
	   error.  This will ultimately cause lpfc_free_scsi_buf to be called
	   which will handle the rest of the cleanup. */
	dma_unmap_sg(&phba->pcidev->dev, sgel_p, use_sg, datadir);
	return 1;
}

static uint32_t
lpfc_os_fcp_err_handle(struct lpfc_scsi_buf *lpfc_cmd,
		       lpfc_xlat_err_t *presult)
{
	struct scsi_cmnd *cmnd = lpfc_cmd->pCmd;
	struct fcp_cmnd *fcpcmd = lpfc_cmd->fcp_cmnd;
	struct fcp_rsp *fcprsp = lpfc_cmd->fcp_rsp;
	struct lpfc_hba *phba = lpfc_cmd->scsi_hba;
	IOCB_t *iocb = &lpfc_cmd->cur_iocbq.iocb;
	uint32_t rsplen = 0;
	/* Make sure presult->host_status is identically DID_OK and scsi_status
	 * is identically 0.  The driver alters this value later on an as-needed
	 * basis.
	 */
	uint32_t scsi_status = 0;
	presult->host_status = DID_OK;

	/*
	 *  If this is a task management command, there is no
	 *  scsi packet associated with it.  Return here.
	 */
	if ((cmnd == 0) || (fcpcmd->fcpCntl2))
		return scsi_status;

	/* FCP cmd failed: RSP */
	lpfc_printf_log(phba,
			KERN_WARNING,
			LOG_FCP,
			"%d:0730 FCP command failed: RSP "
			"Data: x%x x%x x%x x%x x%x x%x\n",
			phba->brd_no,
			fcprsp->rspStatus2, fcprsp->rspStatus3,
			be32_to_cpu(fcprsp->rspResId),
			be32_to_cpu(fcprsp->rspSnsLen),
			be32_to_cpu(fcprsp->rspRspLen), fcprsp->rspInfo3);

	if (fcprsp->rspStatus2 & RSP_LEN_VALID) {
		rsplen = be32_to_cpu(fcprsp->rspRspLen);
		if ((rsplen != 0 && rsplen != 4 && rsplen != 8) ||
			 (fcprsp->rspInfo3 != RSP_NO_FAILURE)) {
				presult->host_status = DID_ERROR;
				scsi_status = fcprsp->rspStatus3;
				return scsi_status;
		}
	}

	/* if there's sense data, let's copy it back */
	if ((fcprsp->rspStatus2 & SNS_LEN_VALID) && (fcprsp->rspSnsLen != 0)) {
		uint32_t snslen = be32_to_cpu(fcprsp->rspSnsLen);

		/* then we return this sense info in the sense buffer for this
		   cmd */
		if (snslen > SCSI_SENSE_BUFFERSIZE)
			snslen = SCSI_SENSE_BUFFERSIZE;

		memcpy(cmnd->sense_buffer,
		       (((uint8_t *) & fcprsp->rspInfo0) + rsplen), snslen);
	}

	/*
	 * In the Tape Env., there is an early WARNNING right before EOM without
	 * data xfer error. We should set b_resid to be 0 before we check all
	 * other cases.
	 */

	cmnd->resid = 0;

	if (fcprsp->rspStatus2 & (RESID_UNDER | RESID_OVER)) {
		if (fcprsp->rspStatus2 & RESID_UNDER) {
			cmnd->resid = be32_to_cpu(fcprsp->rspResId);

			/* FCP Read Underrun, expected <len>, residual
			   <resid> */
			lpfc_printf_log(phba,
					KERN_INFO,
					LOG_FCP,
					"%d:0716 FCP Read Underrun, "
					"expected %d, residual %d "
					"Data: x%x x%x x%x\n",
					phba->brd_no,
					be32_to_cpu(fcpcmd->fcpDl),
					cmnd->resid, iocb->un.fcpi.fcpi_parm,
					cmnd->cmnd[0], cmnd->underflow);

			switch (cmnd->cmnd[0]) {
			case TEST_UNIT_READY:
			case REQUEST_SENSE:
			case INQUIRY:
			case RECEIVE_DIAGNOSTIC:
			case READ_CAPACITY:
			case READ_DEFECT_DATA:
			case MDAC_DIRECT_CMD:
				/* No error */
				return scsi_status;
			default:
				if (!(fcprsp->rspStatus2 & SNS_LEN_VALID) &&
				    ((cmnd->request_bufflen - cmnd->resid) <
				     cmnd->underflow)) {

					/* FCP command <cmd> residual underrun
					   converted to error */
					lpfc_printf_log(phba,
						KERN_INFO,
						LOG_FCP,
						"%d:0717 FCP command x%x "
						"residual underrun converted "
						"to error Data: x%x x%x x%x\n",
						phba->brd_no,
						cmnd->cmnd[0],
						cmnd->request_bufflen,
						cmnd->resid,
						cmnd->underflow);

					presult->host_status = DID_ERROR;
					scsi_status = fcprsp->rspStatus3;

					/* Queue Full handled below */
					if (scsi_status !=
					    SAM_STAT_TASK_SET_FULL)
						return (scsi_status);

				}
			}
		}
	} else {
		if ((lpfc_cmd->datadir == DMA_FROM_DEVICE)
		    && iocb->un.fcpi.fcpi_parm)
		{
			/*
			 * This is ALWAYS a readcheck error!!
			 * Give Check Condition priority over Read Check
			 */

			if (fcprsp->rspStatus3 != SAM_STAT_BUSY) {
				if (fcprsp->rspStatus3 !=
				    SAM_STAT_CHECK_CONDITION) {
					/* FCP Read Check Error */
					lpfc_printf_log(phba,
						KERN_WARNING,
						LOG_FCP,
						"%d:0734 FCP Read Check Error "
						"Data: x%x x%x x%x x%x\n",
						phba->brd_no,	
						be32_to_cpu(fcpcmd->
							    fcpDl),
						be32_to_cpu(fcprsp->
							    rspResId),
						iocb->un.fcpi.fcpi_parm,
						cmnd->cmnd[0]);

					presult->host_status = DID_ERROR;
					cmnd->resid = cmnd->request_bufflen;
					scsi_status = fcprsp->rspStatus3;
					return scsi_status;
				}

				/* FCP Read Check Error with Check Condition */
				lpfc_printf_log(phba,
						KERN_WARNING,
						LOG_FCP | LOG_CHK_COND,
						"%d:0735 FCP Read Check Error "
						"with Check Condition "
						"Data: x%x x%x x%x x%x\n",
						phba->brd_no,
						be32_to_cpu(fcpcmd->fcpDl),
						be32_to_cpu(fcprsp->rspResId),
						iocb->un.fcpi.fcpi_parm,
						cmnd->cmnd[0]);
			}
		}
	}

	scsi_status = fcprsp->rspStatus3;

	switch (scsi_status) {
	case SAM_STAT_TASK_SET_FULL:
		presult->host_status = DID_OK;	/* clear 07 */
		break;

	case SAM_STAT_BUSY:
		presult->host_status = DID_BUS_BUSY;
		scsi_status = (uint32_t) (fcprsp->rspStatus3);
		break;

	case SAM_STAT_CHECK_CONDITION:
		{
			uint32_t i = be32_to_cpu(fcprsp->rspRspLen);
			uint32_t *lp = (uint32_t*)
				       (((uint8_t *) & fcprsp->rspInfo0) + i);
			uint32_t cc = be32_to_cpu((lp[3])
						  & be32_to_cpu(0xFFFF0000));

			/* <ASC ASCQ> Check condition received */
			lpfc_printf_log(phba,
					KERN_INFO,
					LOG_FCP | LOG_CHK_COND,
					"%d:0737 x%x Check condition received "
					"Data: x%x\n",
					phba->brd_no,
					cc,
					*lp);
		}
		break;
	}

	return scsi_status;
}

static void
lpfc_os_return_scsi_cmd(struct lpfc_hba * phba, struct lpfc_scsi_buf * lpfc_cmd)
{
	struct scsi_cmnd *cmd = lpfc_cmd->pCmd;
	lpfc_xlat_err_t resultdata, *presult;
	uint32_t scsi_status = 0;

	if (lpfc_cmd->status >= IOSTAT_CNT)
		lpfc_cmd->status = IOSTAT_DEFAULT;

	presult = &lpfc_iostat_tbl[lpfc_cmd->status];
	if (presult->iocb_status != IOSTAT_SUCCESS) {
		lpfc_printf_log(phba, KERN_WARNING, LOG_FCP,
				"%d:0729 FCP cmd x%x failed <%d/%d> status: "
				"x%x result: x%x Data: x%x x%x\n",
				phba->brd_no, cmd->cmnd[0], cmd->device->id,
				cmd->device->lun, lpfc_cmd->status,
				lpfc_cmd->result, lpfc_cmd->IOxri,
				lpfc_cmd->cur_iocbq.iocb.ulpIoTag);

		if (presult->action_flag & LPFC_FCPRSP_ERROR) {
			/* Call FCP RSP handler to determine result */
			presult = &resultdata;
			scsi_status = lpfc_os_fcp_err_handle(lpfc_cmd, presult);
		}
	}

	cmd->result = ScsiResult(presult->host_status, scsi_status);
	if (cmd->result || lpfc_cmd->fcp_rsp->rspSnsLen) {
		struct lpfc_target *target = cmd->device->hostdata;
		uint32_t *lp = (uint32_t *)cmd->sense_buffer;

		lpfc_printf_log(phba, KERN_INFO, LOG_FCP,
				"%d:0710 Iodone <%d/%d> error x%x "
				"SNS x%x x%x Data: x%x x%x\n",
				phba->brd_no, cmd->device->id,
				cmd->device->lun, cmd->result,
				*lp, *(lp + 3), cmd->retries, cmd->resid);

		if (cmd->result &&
		   (target->pnode) &&
		   (target->pnode->nlp_state != NLP_STE_MAPPED_NODE)) {
			cmd->result = ScsiResult(DID_BUS_BUSY, SAM_STAT_BUSY);
		}
	}

	if (cmd->use_sg) {
		dma_unmap_sg(&phba->pcidev->dev,
			     cmd->request_buffer,
			     cmd->use_sg,
			     lpfc_cmd->datadir);
	} else if (cmd->request_bufflen && lpfc_cmd->nonsg_phys) {
		dma_unmap_single(&phba->pcidev->dev,
				 lpfc_cmd->nonsg_phys,
				 cmd->request_bufflen,
				 lpfc_cmd->datadir);
	}

	lpfc_free_scsi_buf(lpfc_cmd);

	cmd->host_scribble = (unsigned char *) NULL;
	cmd->scsi_done(cmd);
}

static void
lpfc_scsi_cmd_iocb_cmpl(struct lpfc_hba * phba, struct lpfc_iocbq * pIocbIn,
			struct lpfc_iocbq * pIocbOut)
{
	struct lpfc_scsi_buf *pcmd = (struct lpfc_scsi_buf *) pIocbIn->context1;
	struct lpfc_target *target = pcmd->target;

	pcmd->result = pIocbOut->iocb.un.ulpWord[4];
	pcmd->status = pIocbOut->iocb.ulpStatus;
	pcmd->IOxri = pIocbOut->iocb.ulpContext;

	if (pcmd->status == IOSTAT_LOCAL_REJECT &&
	    (pcmd->result & IOERR_DRVR_MASK))
		pcmd->status = IOSTAT_DRIVER_REJECT;

	if (pcmd->status) 	
		target->errorcnt++;

	target->iodonecnt++;

	pci_dma_sync_single_for_cpu(phba->pcidev, pcmd->dma_ext.phys,
			LPFC_BPL_SIZE, PCI_DMA_FROMDEVICE);

	lpfc_os_return_scsi_cmd(phba, pcmd);
}

static int
lpfc_scsi_prep_task_mgmt_cmd(struct lpfc_hba *phba,
			     struct lpfc_scsi_buf *lpfc_cmd,
			     uint8_t task_mgmt_cmd)
{

	struct lpfc_sli *psli;
	struct lpfc_iocbq *piocbq;
	IOCB_t *piocb;
	struct fcp_cmnd *fcp_cmnd;
	struct scsi_device *scsi_dev = lpfc_cmd->pCmd->device;
	struct lpfc_target *target = scsi_dev->hostdata;
	struct lpfc_nodelist *ndlp = target->pnode;

	if ((ndlp == 0) || (ndlp->nlp_state != NLP_STE_MAPPED_NODE)) { 
		return 0;
	}

	/* allocate an iocb command */
	psli = &phba->sli;
	piocbq = &(lpfc_cmd->cur_iocbq);
	piocb = &piocbq->iocb;


	fcp_cmnd = lpfc_cmd->fcp_cmnd;
	putLunHigh(fcp_cmnd->fcpLunMsl, 0);
	putLunLow(fcp_cmnd->fcpLunLsl, lpfc_cmd->pCmd->device->lun)
	fcp_cmnd->fcpCntl2 = task_mgmt_cmd;

	piocb->ulpIoTag =
	    lpfc_sli_next_iotag(phba, &phba->sli.ring[psli->fcp_ring]);
	piocb->ulpCommand = CMD_FCP_ICMND64_CR;

	piocb->ulpContext = ndlp->nlp_rpi;
	if (ndlp->nlp_fcp_info & NLP_FCP_2_DEVICE) {
		piocb->ulpFCP2Rcvy = 1;
	}
	piocb->ulpClass = (ndlp->nlp_fcp_info & 0x0f);

	/* ulpTimeout is only one byte */
	if (lpfc_cmd->timeout > 0xff) {
		/*
		 * Do not timeout the command at the firmware level.
		 * The driver will provide the timeout mechanism.
		 */
		piocb->ulpTimeout = 0;
	} else {
		piocb->ulpTimeout = lpfc_cmd->timeout;
	}

	lpfc_cmd->target = target;

	switch (task_mgmt_cmd) {
	case LUN_RESET:
		/* Issue LUN Reset to TGT <num> LUN <num> */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_FCP,
				"%d:0703 Issue LUN Reset to TGT %d LUN %d "
				"Data: x%x x%x\n",
				phba->brd_no,
				scsi_dev->id, scsi_dev->lun,
				ndlp->nlp_rpi, ndlp->nlp_rflag);

		break;
	case ABORT_TASK_SET:
		/* Issue Abort Task Set to TGT <num> LUN <num> */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_FCP,
				"%d:0701 Issue Abort Task Set to TGT %d LUN %d "
				"Data: x%x x%x\n",
				phba->brd_no,
				scsi_dev->id, scsi_dev->lun,
				ndlp->nlp_rpi, ndlp->nlp_rflag);

		break;
	case TARGET_RESET:
		/* Issue Target Reset to TGT <num> */
		lpfc_printf_log(phba,
				KERN_INFO,
				LOG_FCP,
				"%d:0702 Issue Target Reset to TGT %d "
				"Data: x%x x%x\n",
				phba->brd_no,
				scsi_dev->id, ndlp->nlp_rpi,
				ndlp->nlp_rflag);
		break;
	}

	return (1);
}

static int
lpfc_scsi_tgt_reset(struct lpfc_scsi_buf * lpfc_cmd, struct lpfc_hba * phba,
		     struct lpfc_target *ptarget)
{
	struct lpfc_iocbq *piocbq, *piocbqrsp;
	struct lpfc_sli *psli;
	int ret, err_cnt = 0;

	ret = lpfc_scsi_prep_task_mgmt_cmd(phba, lpfc_cmd, TARGET_RESET);
	if (!ret)
		return FAILED;

	lpfc_cmd->scsi_hba = phba;
	psli = &phba->sli;
	piocbq = &lpfc_cmd->cur_iocbq;
	piocbqrsp = mempool_alloc(phba->iocb_mem_pool,
				  GFP_ATOMIC);
	if (!piocbqrsp)
		return FAILED;

	memset(piocbqrsp, 0, sizeof (struct lpfc_iocbq));

	piocbq->iocb_flag |= LPFC_IO_POLL;

	ret = lpfc_sli_issue_iocb_wait_high_priority(phba,
		     &phba->sli.ring[psli->fcp_ring],
		     piocbq, SLI_IOCB_HIGH_PRIORITY,
		     piocbqrsp,
		     lpfc_cmd->timeout);
	if (ret != IOCB_SUCCESS) {
		lpfc_cmd->status = IOSTAT_DRIVER_REJECT;
	} else {
		lpfc_cmd->result = piocbqrsp->iocb.un.ulpWord[4];
		lpfc_cmd->status = piocbqrsp->iocb.ulpStatus;
		if (lpfc_cmd->status == IOSTAT_LOCAL_REJECT &&
			(lpfc_cmd->result & IOERR_DRVR_MASK))
				lpfc_cmd->status = IOSTAT_DRIVER_REJECT;
	}

	/* 
	 * Flush the target because the HBA does not guarantee to have 
	 * all outstanding IOs terminated following the target reset.
	 */
	err_cnt = lpfc_sli_abort_iocb_tgt(phba,
					  &phba->sli.ring[phba->sli.fcp_ring],
					  ptarget->scsi_id);
	if (err_cnt) {
		lpfc_printf_log(phba, KERN_INFO, LOG_FCP,
				"%d:0715 Abort Target %d IOCB err_cnt: %d\n",
				phba->brd_no, ptarget->scsi_id, err_cnt);
	}

	/* Done with piocbqrsp, return to free list. */
	mempool_free(piocbqrsp, phba->iocb_mem_pool);
	
	return ret;
}


int
lpfc_reset_bus_handler(struct scsi_cmnd *cmnd)
{
	struct Scsi_Host *shost = cmnd->device->host;
	struct lpfc_hba *phba = (struct lpfc_hba *)shost->hostdata[0];
	struct lpfc_target *target;
	int ret = FAILED, i, err_count = 0;
	struct scsi_cmnd *lnx_cmd;
	struct scsi_device *scsi_dev;
	struct lpfc_scsi_buf * lpfc_cmd;

	lpfc_cmd = lpfc_get_scsi_buf(phba, GFP_ATOMIC);
	if (!lpfc_cmd)
		goto out;

	lnx_cmd = kmalloc(sizeof(struct scsi_cmnd), GFP_ATOMIC);
	if (!lnx_cmd)
		goto out_free_scsi_buf;

	scsi_dev = kmalloc(sizeof(struct scsi_device), GFP_ATOMIC);
	if (!scsi_dev)
		goto out_free_lnx_cmd;

	/* The lpfc_cmd storage is reused.  Set all loop invariants. */
	lpfc_cmd->timeout = cmnd->timeout;
	if (!cmnd->timeout)
		 lpfc_cmd->timeout = LPFC_DRVR_TIMEOUT;
 
	lpfc_cmd->pCmd			= lnx_cmd;
	lpfc_cmd->scsi_hba		= phba;
	lpfc_cmd->pCmd->device		= scsi_dev;
	lpfc_cmd->pCmd->device->channel = 0;
	lpfc_cmd->pCmd->device->lun	= 0;

	/* 
	 * Since the driver manages a single bus device, reset all
	 * targets known to the driver.  Should any target reset
	 * fail, this routine returns failure to the midlayer.
	 */
	for (i = 0; i < MAX_FCP_TARGET; i++) {
		target = phba->device_queue_hash[i];
		if (!target)
			continue;

		lpfc_cmd->pCmd->device->id = i;
		lpfc_cmd->pCmd->device->hostdata = target;
		ret = lpfc_scsi_tgt_reset(lpfc_cmd, phba, target);
		if (ret != IOCB_SUCCESS) {
			lpfc_printf_log(phba, KERN_INFO, LOG_FCP,
				"%d:0713 Bus Reset on target %d failed\n",
				phba->brd_no, target->scsi_id);
			err_count++;
		}
	}

out_free_lnx_cmd:
	kfree(lnx_cmd);
out_free_scsi_buf:
	lpfc_free_scsi_buf(lpfc_cmd);
	lpfc_printf_log(phba,
			KERN_ERR,
			LOG_FCP,
			"%d:0714 SCSI layer issued Bus Reset Data: x%x\n",
			phba->brd_no, ret);
out:
	if (!err_count)
		ret = SUCCESS;
	return ret;
}


int
lpfc_queuecommand(struct scsi_cmnd *cmnd, void (*done) (struct scsi_cmnd *))
{
	struct lpfc_hba *phba =
		(struct lpfc_hba *) cmnd->device->host->hostdata[0];
	struct lpfc_sli *psli = &phba->sli;
	struct lpfc_target *targetp = cmnd->device->hostdata;
	struct lpfc_nodelist *ndlp = targetp->pnode;
	struct lpfc_iocbq *piocbq;
	struct lpfc_scsi_buf *lpfc_cmd;
	IOCB_t *piocb;
	int err = 0;

	if (!ndlp) {
		cmnd->result = ScsiResult(DID_NO_CONNECT, 0);
		goto out_fail_command;
	}

	if (ndlp->nlp_state != NLP_STE_MAPPED_NODE) {
		if ((ndlp->nlp_state == NLP_STE_UNMAPPED_NODE) ||
		    (ndlp->nlp_state == NLP_STE_UNUSED_NODE)) {
			cmnd->result = ScsiResult(DID_NO_CONNECT, 0);
			goto out_fail_command;
		}
		/* 
		 * In this case, the device may be on the way back,
		 * and the driver wants to notify the midlayer
		 * to retry.
		 */
		goto out_host_busy;
	}

	lpfc_cmd = lpfc_get_scsi_buf(phba, GFP_ATOMIC);
	if (!lpfc_cmd)
		goto out_host_busy;

	/*
	 * Store the midlayer's command structure for the completion phase
	 * and complete the command initialization.
	 */
	cmnd->scsi_done = done;
	cmnd->host_scribble = (unsigned char *)lpfc_cmd;

	lpfc_cmd->target = targetp;
	lpfc_cmd->cmd_cmpl = lpfc_os_return_scsi_cmd;
	lpfc_cmd->timeout = 0;
	lpfc_cmd->pCmd = cmnd;
	putLunHigh(lpfc_cmd->fcp_cmnd->fcpLunMsl, lpfc_cmd->pCmd->device->lun);
	putLunLow(lpfc_cmd->fcp_cmnd->fcpLunLsl, lpfc_cmd->pCmd->device->lun);

	err = lpfc_os_prep_io(phba, lpfc_cmd);
	if (err) {
		/*
		 * The driver could not complete the sg and bpl initialization.
		 * This is a hard failure.
		 */
		lpfc_free_scsi_buf(lpfc_cmd);
		return 1;
	}

	piocbq = &(lpfc_cmd->cur_iocbq);
	piocb = &piocbq->iocb;
	piocb->ulpTimeout = lpfc_cmd->timeout;
	piocbq->context1 = lpfc_cmd;
	piocbq->iocb_cmpl = lpfc_scsi_cmd_iocb_cmpl;

	piocbq->iocb.ulpContext = targetp->pnode->nlp_rpi;
	if (targetp->pnode->nlp_fcp_info & NLP_FCP_2_DEVICE) {
		piocbq->iocb.ulpFCP2Rcvy = 1;
	}

	piocbq->iocb.ulpClass = (targetp->pnode->nlp_fcp_info & 0x0f);
	piocbq->iocb.ulpIoTag = lpfc_sli_next_iotag(phba,
						   &psli->ring[psli->fcp_ring]);

	pci_dma_sync_single_for_device(phba->pcidev, lpfc_cmd->dma_ext.phys,
			LPFC_SCSI_DMA_EXT_SIZE, PCI_DMA_TODEVICE);

	err = lpfc_sli_issue_iocb(phba, &phba->sli.ring[psli->fcp_ring], piocbq,
				 SLI_IOCB_RET_IOCB);
	if (err)
		goto out_host_busy_free_buf;

	targetp->qcmdcnt++;
	return 0;

 out_host_busy_free_buf:
	lpfc_free_scsi_buf(lpfc_cmd);
 out_host_busy:
	return SCSI_MLQUEUE_HOST_BUSY;

 out_fail_command:
	done(cmnd);
	return 0;
}

int
lpfc_reset_lun_handler(struct scsi_cmnd *cmnd)
{
	struct Scsi_Host *shost = cmnd->device->host;
	struct lpfc_hba *phba = (struct lpfc_hba *)shost->hostdata[0];
	struct lpfc_sli *psli = &phba->sli;
	struct lpfc_scsi_buf *lpfc_cmd;
	struct lpfc_iocbq *piocbq, *piocbqrsp = NULL;
	int ret = FAILED;

	lpfc_cmd = lpfc_get_scsi_buf(phba, GFP_ATOMIC);
	if (!lpfc_cmd)
		goto out;

	lpfc_cmd->pCmd = cmnd;
	lpfc_cmd->timeout = 60; /* set command timeout to 60 seconds */
	lpfc_cmd->scsi_hba = phba;
	
	ret = lpfc_scsi_prep_task_mgmt_cmd(phba, lpfc_cmd, LUN_RESET);
	if (!ret)
		goto out_free_scsi_buf;
	
	piocbq = &lpfc_cmd->cur_iocbq;

	/* get a buffer for this IOCB command response */
	piocbqrsp = mempool_alloc(phba->iocb_mem_pool, GFP_ATOMIC);
	if(!piocbqrsp)
		goto out_free_scsi_buf;

	memset(piocbqrsp, 0, sizeof (struct lpfc_iocbq));

	piocbq->iocb_flag |= LPFC_IO_POLL;
	piocbq->iocb_cmpl = lpfc_sli_wake_iocb_high_priority;

	ret = lpfc_sli_issue_iocb_wait_high_priority(phba,
		     &phba->sli.ring[psli->fcp_ring],
		     piocbq, SLI_IOCB_USE_TXQ,
		     piocbqrsp, 60);
	if (ret == IOCB_SUCCESS)
		ret = SUCCESS;
	
	lpfc_cmd->result = piocbqrsp->iocb.un.ulpWord[4];
	lpfc_cmd->status = piocbqrsp->iocb.ulpStatus;
	if (lpfc_cmd->status == IOSTAT_LOCAL_REJECT)
		if (lpfc_cmd->result & IOERR_DRVR_MASK)
			lpfc_cmd->status = IOSTAT_DRIVER_REJECT;

	/* flush outstanding commands on the lun */
	lpfc_sli_abort_iocb_lun(phba,
				&phba->sli.ring[phba->sli.fcp_ring],
				cmnd->device->id,
				cmnd->device->lun);

	mempool_free(piocbqrsp, phba->iocb_mem_pool);

out_free_scsi_buf:
	lpfc_printf_log(phba, KERN_ERR, LOG_FCP,
			"%d:713 SCSI layer issued target reset "
			"Data: x%x x%x x%x x%x x%x\n",
			phba->brd_no, ret, lpfc_cmd->pCmd->device->id,
			lpfc_cmd->pCmd->device->lun, lpfc_cmd->status,
			lpfc_cmd->result);
	lpfc_free_scsi_buf(lpfc_cmd);
out:
	return ret;
}


int
lpfc_abort_handler(struct scsi_cmnd *cmnd)
{
	struct Scsi_Host *shost = cmnd->device->host;
	struct lpfc_hba *phba = (struct lpfc_hba *)shost->hostdata[0];
	struct lpfc_sli *psli = &phba->sli;
	struct lpfc_sli_ring *pring = &psli->ring[psli->fcp_ring];
	struct lpfc_iocbq *iocb, *next_iocb, *abtsiocbp;
	struct lpfc_scsi_buf *lpfc_cmd;
	IOCB_t *cmd,*icmd;
	int errcnt = 0, ret = FAILED;

	lpfc_cmd = (struct lpfc_scsi_buf *)cmnd->host_scribble;
	if (!lpfc_cmd)
		goto out;

	/* set command timeout to 60 seconds */
	lpfc_cmd->timeout = 60;

	list_for_each_entry_safe(iocb, next_iocb, &pring->txq, list) {
		cmd = &iocb->iocb;
		if (iocb->context1 != lpfc_cmd)
			continue;

		list_del_init(&iocb->list);
		pring->txq_cnt--;
		if (!iocb->iocb_cmpl) {
			mempool_free(iocb, phba->iocb_mem_pool);
			continue;
		} 
		cmd->ulpStatus = IOSTAT_LOCAL_REJECT;
		cmd->un.ulpWord[4] = IOERR_SLI_ABORTED;
		(iocb->iocb_cmpl)(phba, iocb, iocb);
	}

	/* Next check the txcmplq */
	list_for_each_entry_safe(iocb, next_iocb, &pring->txcmplq, list) {
		if (iocb->context1 != lpfc_cmd)
			continue;

		cmd = &iocb->iocb;
		/* issue ABTS for this IOCB based on iotag */
		abtsiocbp = mempool_alloc(phba->iocb_mem_pool, GFP_ATOMIC);
		if (!abtsiocbp) {
			errcnt++;
			continue;
		}
		memset(abtsiocbp, 0, sizeof (struct lpfc_iocbq));

		icmd = &abtsiocbp->iocb;
		icmd->un.acxri.abortType = ABORT_TYPE_ABTS;
		icmd->un.acxri.abortContextTag = cmd->ulpContext;
		icmd->un.acxri.abortIoTag = cmd->ulpIoTag;

		icmd->ulpLe = 1;
		icmd->ulpClass = cmd->ulpClass;
		if (phba->hba_state >= LPFC_LINK_UP)
			icmd->ulpCommand = CMD_ABORT_XRI_CN;
		else
			icmd->ulpCommand = CMD_CLOSE_XRI_CN;

		/* set up an iotag  */
		icmd->ulpIoTag = lpfc_sli_next_iotag(phba, pring);

		if (lpfc_sli_issue_iocb(phba, pring, abtsiocbp,
					SLI_IOCB_USE_TXQ) == IOCB_ERROR) {
			mempool_free(abtsiocbp, phba->iocb_mem_pool);
			errcnt++;
			continue;
		}

		/*
		 * The rsp ring completion will remove IOCB from txcmplq when
		 * abort is read by HBA.
		 */
	}
	ret = (errcnt? FAILED: SUCCESS);

	lpfc_printf_log(phba,
			KERN_ERR,
			LOG_FCP,
			"%d:0749 SCSI layer issued abort device "
			"Data: x%x x%x x%x x%x x%x\n",
			phba->brd_no, ret,
			lpfc_cmd->pCmd->device->id,
			lpfc_cmd->pCmd->device->lun,
			lpfc_cmd->status, lpfc_cmd->result);

 out:			
	return ret;
}

void
lpfc_target_unblock(struct lpfc_hba *phba, struct lpfc_target *targetp)
{
	struct lpfc_nodelist *ndlp = targetp->pnode;

	/* Resume Target scsi id <sid> to NPort <nlp_DID> */
	lpfc_printf_log(phba, KERN_INFO, LOG_DISCOVERY | LOG_FCP,
			"%d:0258 Unblocking IO to Target scsi id x%x to "
			"NPort x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			ndlp->nlp_sid, ndlp->nlp_DID,
			ndlp->nlp_flag, ndlp->nlp_state, ndlp->nlp_rpi);
	/* fc_target_unblock(targetp->starget); */
}

int
lpfc_target_block(struct lpfc_hba *phba, struct lpfc_target *targetp)
{
	int err = 0;
	struct lpfc_nodelist *ndlp = targetp->pnode;

	/* Suspend Target scsi id <sid> to NPort <nlp_DID> */
	lpfc_printf_log(phba, KERN_INFO, LOG_DISCOVERY | LOG_FCP,
			"%d:0259 Blocking IO to Target scsi id x%x to "
			"NPort x%x Data: x%x x%x x%x\n",
			phba->brd_no,
			ndlp->nlp_sid, ndlp->nlp_DID,
			ndlp->nlp_flag, ndlp->nlp_state, ndlp->nlp_rpi);
	/* err = fc_target_block(targetp->starget); */
	return err;
}

int
lpfc_target_remove(struct lpfc_hba *phba, struct lpfc_target *targetp)
{
	/* Remove Target scsi id <sid> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY | LOG_FCP,
			"%d:0260 Remove Target scsi id x%x\n",
			phba->brd_no,
			targetp->scsi_id);
	/* call fc_target_unblock */
	/* call scsi_scan_host_select, or equiv, for id */
	return (0);
}

int
lpfc_target_add(struct lpfc_hba *phba, struct lpfc_target *targetp)
{
	/* Add Target scsi id <sid> */
	lpfc_printf_log(phba,
			KERN_INFO,
			LOG_DISCOVERY | LOG_FCP,
			"%d:0261 Add Target scsi id x%x\n",
			phba->brd_no,
			targetp->scsi_id);
	/* call scsi_scan_host_select, or equiv, for id */
	return (0);
}

